/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#define _CONDOR_ALLOW_OPEN
#include "condor_common.h"
#include "condor_debug.h"
#include "condor_daemon_core.h"
#include "dap_server.h"
#include "subsystem_info.h"
#include "stork-lm.h"

DECL_SUBSYSTEM("STORK", SUBSYSTEM_TYPE_DAEMON);  // used by Daemon Core

// Command line arguments are read into these variables
char *logfilename = "storkserver.log";
char *xmllogfilename = NULL;
char *userlogfilename = NULL;
char *clientagenthost = NULL;
// Timers
int IdleJobMonitorInterval;
int IdleJobMonitorTid;
int HungJobMonitorInterval;
int HungJobMonitorTid;
int RescheduledJobMonitorInterval;
int RescheduledJobMonitorTid;
int LowWaterTid;

extern int transfer_dap_reaper_id, reserve_dap_reaper_id;
extern int release_dap_reaper_id;
extern int requestpath_dap_reaper_id;
extern int low_water_reaper_id;
extern int estimate_dap_reaper_id;

// SERVER INITIALIZATION
// =====================
// Handle parsing of command line arguments and start various 

// Simply print the usage information for this command.
void print_usage() {
	fprintf(stderr,
		"==============================================================\n"
		"USAGE: stork_server\n"
		"   [ -t           ] // output to stdin\n"
		"   [ -p           ] // port on which to run Stork Server\n"
		"   [ -help        ] // stork help screen\n"
		"   [ -Config      ] // stork config file\n"
		"   [ -Serverlog   ] // stork server log in ClassAds\n"
		"   [ -Xmllog      ] // stork server log in XML format\n"
		"==============================================================\n");
}

// Parse command line arguments
void parse_arguments(int argc, char **argv) {
	for (int i = 1; i < argc; i++) {
		// Parse out Serverlog argument
		if (!strcmp("-Serverlog", argv[i])) {
			if (++i >= argc) {
				dprintf(D_ALWAYS, "No daplog file specified!\n");
				print_usage();
				exit(EXIT_FAILURE);
			} logfilename = argv[i];
		}
		// Parse out Xmllog argument
		else if (!strcmp("-Xmllog", argv[i])) {
			if (++i >= argc) {
				dprintf(D_ALWAYS, "No xmllog file specified!\n");
				print_usage();
				exit(EXIT_FAILURE);
			} xmllogfilename = argv[i];
		}
		// Skip over -p argument so it can be handled elsewhere
		else if (!strcmp("-p", argv[i])) i++;
		// Print usage and exit successfully if -help passed
		else if (!strcmp("-help", argv[i])) {
			print_usage();
			exit(EXIT_SUCCESS);
		}
		// Exit if passed invalid argument
		else {
			print_usage();
			exit(EXIT_FAILURE);
		}
	}
}

int main_init(int argc, char **argv) {
	parse_arguments(argc, argv);
	
	// Dump the configuration
	dprintf(D_ALWAYS, "DaP log file: %s\n", "XML log file: %s\n",
	        logfilename, xmllogfilename);

	// Submit stat report
	if (stat_report_enabled()) {
		dprintf(D_ALWAYS, "Submitting anonymous usage report to %s\n", 
		        stat_server_host());
		send_stat_report();
	}

	// Register Stork commands
	daemonCore->Register_Command(
		STORK_SUBMIT, "STORK_SUBMIT",
		(CommandHandler) &handle_stork_submit,
		"handle_stork_submit", NULL, WRITE);

	daemonCore->Register_Command(
		STORK_STATUS, "STORK_STATUS",
		(CommandHandler) &handle_stork_status,
		"handle_stork_status", NULL, WRITE);

	daemonCore->Register_Command(
		STORK_LIST, "STORK_LIST",
		(CommandHandler) &handle_stork_list,
		"handle_stork_list", NULL, WRITE);

	daemonCore->Register_Command(
		STORK_REMOVE, "STORK_REMOVE",
		(CommandHandler) &handle_stork_remove,
		"handle_stork_remove", NULL, WRITE);
  
	daemonCore->Register_Command(
		STORK_LOG, "STORK_LOG",
		(CommandHandler) &handle_stork_log,
		"handle_stork_log", NULL, WRITE);

	// Activate job monitors
	HungJobMonitorInterval = param_integer(
		"HUNG_JOB_MONITOR",
		HUNG_JOB_MONITOR_DEFAULT,
		HUNG_JOB_MONITOR_MIN);

	HungJobMonitorTid = daemonCore->Register_Timer(
		4,                       // deltawhen
		HungJobMonitorInterval,  // period
		(TimerHandler) regular_check_for_requests_in_process,
		"check_for_requests_in_process");

	IdleJobMonitorInterval =
		param_integer(
			"IDLE_JOB_MONITOR",
			IDLE_JOB_MONITOR_DEFAULT,
			IDLE_JOB_MONITOR_MIN);
	IdleJobMonitorTid = 
		daemonCore->Register_Timer(
			1,                        // deltawhen
			IdleJobMonitorInterval,   // period
			(TimerHandler)call_main,  // event
			"call_main");             // description

	RescheduledJobMonitorInterval =
		param_integer(
			"RESCHEDULED_JOB_MONITOR",
			RESCHEDULED_JOB_MONITOR_DEFAULT,
			RESCHEDULED_JOB_MONITOR_MIN);
	RescheduledJobMonitorTid = 
		daemonCore->Register_Timer(
			7,                              //deltawhen
			RescheduledJobMonitorInterval,  // period
			(TimerHandler)regular_check_for_rescheduled_requests,
			"regular_check_for_rescheduled_requests");

#if SC2005_DEMO
	LowWaterTid =
		daemonCore->Register_Timer(
			1,   // deltawhen
			60,  // period
			(TimerHandler) low_water_timer,
			"low_water_timer");
#endif

	// Register reaper functions
	transfer_dap_reaper_id = daemonCore->Register_Reaper(
		"transfer_dap_reaper",
		(ReaperHandler) transfer_dap_reaper,
		"reaper for transfer DaP");

	reserve_dap_reaper_id = daemonCore->Register_Reaper(
		"reserve_dap_reaper",
		(ReaperHandler)reserve_dap_reaper,
		"reaper for reserve DaP");

	release_dap_reaper_id = daemonCore->Register_Reaper(
		"release_dap_reaper",
		(ReaperHandler)release_dap_reaper,
		"reaper for release DaP");

	requestpath_dap_reaper_id = daemonCore->Register_Reaper(
		"requestpath_dap_reaper",
		(ReaperHandler)requestpath_dap_reaper,
		"reaper for requestpath DaP");

#if SC2005_DEMO
	low_water_reaper_id = daemonCore->Register_Reaper(
		"low_water_reaper",
		(ReaperHandler)low_water_reaper,
		"reaper for low_water");
#endif

	if (!initializations()) {
		DC_Exit(EXIT_FAILURE);
		return FALSE;
	} return TRUE;
}

// SERVER CONTROLS
// ===============
// Signal handlers that cause the server to do certain things like shutdown
// or reload its configuration.

// NOTE: These used to be ints and main_config had an int argument.
// Reconfigure reqex_daemon
void main_config(bool) {
	dprintf(D_ALWAYS, "RECONFIGURING...\n");
	read_config_file();
}

// Shutdown reqex_daemon fast
void main_shutdown_fast() {
	dprintf(D_ALWAYS, "SHUTDOWN FAST...\n");
	terminate(TERMINATE_FAST);
  
	DC_Exit(EXIT_SUCCESS);
}

// Shutdown reqex_daemon gracefully
void main_shutdown_graceful() {
	dprintf(D_ALWAYS, "SHUTDOWN GRACEFULLY...\n");
	terminate(TERMINATE_GRACEFUL);

	DC_Exit(EXIT_SUCCESS);
}

// wtf is this???
void main_pre_dc_init(int argc, char **argv) {
	(void) argc;
	(void) argv;
}

// STORK COMMAND HANDLERS
// ======================
// These take care of the task of responding to requests from clients.

// stork_submit handler (helper)
int handle_stork_submit_more(Service *, Stream *s) {
	return write_requests_to_file((ReliSock*) s);
}

// stork_submit handler
int handle_stork_submit(Service *, int command, Stream *s) {
	dprintf(D_COMMAND, "handle_stork_submit (%d)\n", command);
	
	if (s->type() != Stream::reli_sock ) {
		dprintf(D_ALWAYS, "ERROR: Socket not of type reli_sock!\n");
		return FALSE;
	}

	int rc = write_requests_to_file((ReliSock*) s);
	if (rc == KEEP_STREAM)
		daemonCore->Register_Socket(
			s, "STORK_SUBMIT",
			(SocketHandler) &handle_stork_submit_more,
			"handle_stork_submit", NULL, WRITE);
	return rc;
}

// Is this necessary?
void main_pre_command_sock_init(void) { }

// Handler for stork_q
int handle_stork_list(Service *, int command, Stream *s) {
	dprintf (D_COMMAND, "handle_stork_list (%d)\n", command);

	if (s->type() != Stream::reli_sock ) {
		dprintf (D_ALWAYS, "ERROR: Socket not of type reli_sock!\n");
		return FALSE;
	}

	return list_queue((ReliSock*) s);
}

// Handler for stork_rm
int handle_stork_remove(Service *, int command, Stream *s) {
	dprintf (D_COMMAND, "handle_stork_remove (%d)\n", command);
	
	if (s->type() != Stream::reli_sock) {
		dprintf (D_ALWAYS, "ERROR: Socket not of type reli_sock!\n");
		return FALSE;
	}

	return remove_requests_from_queue((ReliSock*) s);
}

// Handler for stork_status
int handle_stork_status(Service *, int command, Stream *s) {
	dprintf (D_COMMAND, "handle_stork_status (%d)\n", command);
	if (s->type() != Stream::reli_sock ) {
		dprintf (D_ALWAYS, "ERROR: Socket not of type reli_sock!\n");
		return FALSE;
	}

	return send_dap_status_to_client((ReliSock*) s);
}

// Handler for stork_log
int handle_stork_log(Service *, int command, Stream *s) {
	dprintf(D_COMMAND, "handle_stork_log (%d)\n", command);
	if (s->type() != Stream::reli_sock ) {
		dprintf (D_ALWAYS, "ERROR: Socket not of type reli_sock!\n");
		return FALSE;
	}

	return send_stork_log_to_client((ReliSock*) s);
}
