/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "condor_common.h"
#include "dap_constants.h"
#include "dap_utility.h"
#include "dap_error.h"
#include "dap_logger.h"
#include <unistd.h>
#include "dap_url.h"

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <string.h>

#define MAX_STR 512

// mkdir at local filesystem
int file_mkdir(char *dst, int mymode) {
	char *p = dst;

	do {
		p = strchr(p+1, '/');
		if (p) *p = '\0';

		// errno == 17 means that the directory already exists
		// we therefore ignore this error
		if (mkdir(dst, mymode) == -1 && errno != 17) {
			fprintf(stderr, "mkdir %s failed: %s %d\n",
			        dst, strerror(errno), errno);
			return -1;
		}
		if (p) *p = '/';
	} while (p);

	return 0;
}

// Setting stork specific parameters (filesize verify, chksum verify)
void set_stork_parameters(int *fs_verify, int *cs_verify, int *mymode) {
	char *envVal;

	// Set file permissions as provided
	if (envVal = getenv("STORK_SET_PERMISSIONS"))
		*mymode = atoi(envVal);

	// Perform checksum verification after data transfer
	if ((envVal = getenv("STORK_VERIFY_FILESIZE")) &&
	    (!strncasecmp("TRUE", envVal, 4)))
		*fs_verify = 1;

	// Perform checksum verification after data transfer
	if ((envVal = getenv("STORK_VERIFY_CHECKSUM")) &&
	    (!strncasecmp("TRUE", envVal, 4)))
		*cs_verify = 1;
}

// Transfer a file from local disk to local disk
int transfer_from_file_to_file(char *src_file, char *dest_file,
		int filesizeVerify, int checksumVerify, int mymode) {
	int in_fd, out_fd;
	char *buf = NULL;
	long src_filesize = -1, dest_filesize = -1;
	int bytes_read, bytes_written;
	struct stat64 src_filestat, dest_filestat;
	char localchksum[10], localchksum2[10];
	char chksumbuf[10 + strlen(src_file)], chksumbuf2[10 + strlen(dest_file)];
	FILE *p_fp = NULL, *p_fp2 = NULL;

	// start transfer
	buf = (char*) malloc(sizeof(char) * BUF_SIZE);

	// get the src filesize info
	if (!lstat64(src_file, &src_filestat))
		src_filesize = src_filestat.st_size;

	// open the input file for reading
	in_fd = open(src_file, O_RDONLY);

	if (in_fd < 0){
		fprintf(stdout, "Error in opening file : %s\n",src_file);
		free(buf);
		return DAP_ERROR;
	}

	// open the output file for writing
	out_fd = open(dest_file, O_CREAT | O_RDWR, mymode);

	if (out_fd < 0) {
		fprintf(stdout, "Error in opening file : %s\n",dest_file);
		free(buf);
		close(in_fd);
		return DAP_ERROR;
	}

	// transfer file
	while ((bytes_read = read(in_fd, buf, BUF_SIZE)) > 0) {
		bytes_written = write(out_fd, buf, bytes_read);

		if (bytes_written != bytes_read) {
			fprintf(stdout, "Error! read = %d, written = %d\n",
					bytes_read, bytes_written);
			free(buf);
			close(in_fd);
			close(out_fd);
			return DAP_ERROR;
		}
	}

	// get the dest filesize info
	if (!lstat64(dest_file, &dest_filestat))
		dest_filesize = dest_filestat.st_size;

	// file size verification
	if (filesizeVerify && src_filesize != dest_filesize) {
		printf("src and dest file sizes do not match!!\n");
		free(buf);
		close(in_fd);
		close(out_fd);
		return DAP_ERROR;
	}

	// checksum verification
	// TODO: Perhaps we should verify the checksum without using the sum utility
	if (checksumVerify) {
		// calculate the checksum for src file and dest file
		sprintf(chksumbuf,  "sum -s %s", src_file);
		sprintf(chksumbuf2, "sum -s %s", dest_file);
		
		p_fp  = popen(chksumbuf,  "r");
		p_fp2 = popen(chksumbuf2, "r");
			
		if (!p_fp || !p_fp2) {
			fprintf(stderr, "Error: unable to execute sum\n");
			goto fail;
		}
		
		fscanf(p_fp,  "%s", localchksum);
		fscanf(p_fp2, "%s", localchksum2);
		
		// verify source and destination checksum values
		if (atoi(localchksum) != atoi(localchksum2)) {
			fprintf(stderr,
			        "Error: checksum mismatch src: %d, dest: %d... "
					"deleting dest file\n",
			        atoi(localchksum), atoi(localchksum2));
			unlink(dest_file);
			goto fail;
		}
	}

	// otherwise file transferred succesfully
	printf("SUCCESS!\n");
	free(buf);
	return DAP_SUCCESS;

	// Error handling
	fail:
	free(buf);
	if (in_fd)  close(in_fd);
	if (out_fd) close(out_fd);
	if (p_fp)   pclose(p_fp);
	if (p_fp2)  pclose(p_fp2);
	return DAP_ERROR;
}

// Do a transfer from local disk to local disk
int file_to_file(char *src, char *dst) {
	int status;
	int checksumVerify = 0, filesizeVerify = 0, mymode = 0700;

	// Parse source url
	Url fileParsed(src);
	if (!fileParsed.parsed()) {
		fprintf(stderr, "parse src URL '%s' error\n",
		        fileParsed.urlNoPassword());	
		return -1;
	}

	// Parse destination url
	Url fileParsed2(dst);
	if (!fileParsed2.parsed()) {
		fprintf(stderr, "parse dest URL '%s' error\n",
		fileParsed2.urlNoPassword());
		return -1;
	}

	if (!fileParsed.path() || !fileParsed2.path()) {
		fprintf(stderr,
		        "Error: some fields are missing. "
		        "Please recheck your URL and resubmit the job.\n"
		        "src:%s\ndest:%s\n",
		        fileParsed.urlNoPassword(), fileParsed2.urlNoPassword());
		return -1;
	}

	// Obtain Stork Parameters
	set_stork_parameters(&filesizeVerify, &checksumVerify, &mymode);

	// Recursive transfer
	if (fileParsed.dirname() != NULL && fileParsed2.dirname() != NULL &&
	    fileParsed.basename() == NULL && fileParsed2.basename() == NULL) {
		
		// Perform recursive copy of the directory given in the source url
		FILE *infile;
		char path[MAX_STR], dest[MAX_STR], cmd[MAX_STR];

		// Create required directories
		// TODO: No, this is bad.
		sprintf(cmd, "find %s -type d", fileParsed.dirname());
		infile = popen(cmd, "r");
		if (!infile) {
			fprintf(stderr, "Error: Unable to open: tmp-filelist\n");
			return -1;
		}

		while (fgets(path, MAX_STR, infile)) {
			sprintf(dest, "%s%s", fileParsed2.dirname(),
			        path+strlen(fileParsed.dirname()));
			
			status = file_mkdir(dest, mymode);
			if (status < 0) {
			    fprintf(stderr, "Collection creation error, status=%d\n",
				        status);
				return -1;
			}
		} fclose(infile);

		// List all files in the directory recursively and transfer
		sprintf(cmd, "find %s -type f", fileParsed.dirname());
		infile = popen(cmd, "r");

		if (!infile) {
		    fprintf(stderr, "Error: Unable to open tmp-filelist\n");	
			return -1;
		}
		
		while (fgets(path, MAX_STR, infile)) {
			sprintf(dest, "%s%s", fileParsed2.dirname(),
						path+strlen(fileParsed.dirname()));

			status = transfer_from_file_to_file(path, dest, 
						  filesizeVerify, checksumVerify, mymode); 

			if (status < 0) {
				fprintf(stderr, "Error: transfer file"
					" %s to file %s failed!\n",
						path, dest);
				return -1;
				
			}
		} fclose(infile);
	}

	// Wildcard transfer
	else if (fileParsed.basename() && !fileParsed2.basename() &&
			 strchr(fileParsed.basename(), '*')) {

		FILE *infile;
		char path[MAX_STR], dest[MAX_STR], cmd[MAX_STR];

		status = file_mkdir(fileParsed2.dirname(), mymode);
		if (status < 0){
			fprintf(stderr, "Collection creation error, status =%d\n",status);
			return -1;
		}
	  
		sprintf(cmd, "find %s -maxdepth 1 -name \"%s\" -type f", 
		        fileParsed.dirname(), fileParsed.basename());
		infile = popen(cmd, "r");
		if (!infile) {
			fprintf(stderr, "Error, Unable to open pipe:\n");
			return -1;
		}
	
		while (fgets(path, MAX_STR, infile)) {
			sprintf(dest, "%s%s", fileParsed2.dirname(), strrchr(path, '/'));
			status = transfer_from_file_to_file(path, dest,	
					filesizeVerify, checksumVerify, mymode);

			if (status < 0){
				fprintf(stderr, "Error: transfer" 
					"file %s to file %s failed\n",
					path, dest);
				return -1;
			}
		} fclose(infile);
	}

	// Single file transfer
	else if (fileParsed.basename() && !strchr(fileParsed.basename(), '*')) {
		char dest[MAX_STR];
		int len = strlen(fileParsed2.path());
		snprintf(dest, MAX_STR, "%s", fileParsed2.path());

		// create required directories
		status = file_mkdir(fileParsed2.dirname(), mymode);

		if (status < 0) {
			fprintf(stderr, "Collection creation error, status = %d\n", status);
			return -1;
		}

		// use source filename as desination filename if destination file name
		// not given
		if (!fileParsed2.basename())
			snprintf(dest, MAX_STR, "%s/%s",
			         fileParsed2.dirname(), fileParsed.basename());


		// call to transfer function
		status = transfer_from_file_to_file(fileParsed.path(), dest,
						filesizeVerify, checksumVerify, mymode);

		if (status < 0) {
		    fprintf(stderr, "Error: transfer file %s to file % s failed\n",
				fileParsed.path(), dest);
		    return -1;
		}
	}
	
	// Something was wrong with the URLs
	else {
		fprintf(stderr, "please check the submitted urls and try again\n");
		fprintf(stderr, "src:%s dest:%s\n", fileParsed.urlNoPassword(),
						    fileParsed2.urlNoPassword());
		return -1;
	}	

	return 0;
}

int main(int argc, char *argv[]) {
	if (argc < 3) {
		fprintf(stderr,
			"==============================================================\n"
			"USAGE: %s <src_url> <dest_url>\n"
			"==============================================================\n",
			argv[0]
		);
		return EXIT_FAILURE;
	}

	if (!strncmp(argv[1], "file:/", 6) && !strncmp(argv[2], "file:/", 6)) {
		if (file_to_file(argv[1], argv[2]) < 0) {
			fprintf(stderr, "Error: could not transfer file to file\n");
			return EXIT_FAILURE;
		} return EXIT_SUCCESS;
	} else {
		printf("please check the syntax of the file jobs submitted");
		return EXIT_FAILURE;
	}
}
