/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

// Transfers files to GridFTP servers, from GridFTP servers, and between
// GridFTP servers.


// - Use arbitrary length integers (bigints) instead of uintmax_t.
// - Limit or eliminate user-specified sample sizes.

//#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>

#include "condor_common.h"
#include "dap_constants.h"
#include "dap_error.h"
#include "dap_utility.h"
#include "condor_string.h"
#include "condor_config.h"
#include "env.h"
#include "setenv.h"
#include "my_hostname.h"
#include "globus_ftp_client.h"
#include "dap_guc_optimization.h"

static globus_mutex_t glock;
static globus_cond_t cond;
static globus_bool_t done;
static globus_bool_t error = GLOBUS_FALSE;

#define STORK_ENVIRONMENT "STORK_ENVIRONMENT"
#define MAXSIZE 100
#define SIZE 42
#define SIZE_UNKNOWN	((uintmax_t)(-1))
#define GLOBUS_URL_COPY	"globus-url-copy"

// Maximum number of chunks to split a file into for optimization sampling.
#define MAX_CHUNKS 5

void print_usage(char *prog_name) {
	fprintf(stderr,
	        "==============================================================\n"
	        "USAGE: %s <src_url> <dest_url> [<arguments>]\n"
	        "==============================================================\n",
	        prog_name);
}

/* ========================================================================== */
static void done_cb(void *user_arg, globus_ftp_client_handle_t *handle,
		globus_object_t *err) {
	char * tmpstr;

	if (err) {
		tmpstr = globus_object_printable_to_string(err);
		fprintf(stderr, "%s\n", tmpstr);
		error = GLOBUS_TRUE;
		globus_libc_free(tmpstr);
	}
	globus_mutex_lock(&glock);
	done = GLOBUS_TRUE;
	globus_cond_signal(&cond);
	globus_mutex_unlock(&glock);
}

/* ========================================================================== */
// This is the only reason we need the GridFTP external.
// Isn't there a better way?
int get_filesize_from_gridftp(const char *src_url, uintmax_t *filesize) {
	globus_ftp_client_handle_t        handle;
	globus_ftp_client_operationattr_t attr;
	globus_ftp_client_handleattr_t    handle_attr;
	globus_result_t                   res;
	globus_off_t                      size;
	
	globus_module_activate(GLOBUS_FTP_CLIENT_MODULE);
	globus_ftp_client_handleattr_init(&handle_attr);
	globus_ftp_client_operationattr_init(&attr);
	
	globus_mutex_init(&glock, GLOBUS_NULL);
	globus_cond_init(&cond, GLOBUS_NULL);
	globus_ftp_client_handle_init(&handle, &handle_attr);
	done = GLOBUS_FALSE;
	
    res = globus_ftp_client_size(&handle, src_url, &attr, &size, done_cb, 0);
	
	if (res != GLOBUS_SUCCESS) {
		fprintf(stderr, "%s\n",
		        globus_object_printable_to_string(globus_error_get(res)));
		error = GLOBUS_TRUE;
		done = GLOBUS_TRUE;
	}
	
	globus_mutex_lock(&glock);
	
	while (!done) globus_cond_wait(&cond, &glock);
	
	globus_mutex_unlock(&glock);
	globus_ftp_client_handle_destroy(&handle);
	globus_module_deactivate_all();
	
	if (error != GLOBUS_SUCCESS) {
		return DAP_ERROR;
	} else {
		*filesize = size;
		return DAP_SUCCESS;
	}
}
/* ========================================================================== */
int transfer_globus_url_copy(const char *src_url, const char *dest_url,
                             const char *arguments, char *error_str) {
	FILE *pipe = 0;
	char pipecommand[MAXSTR];
	char linebuf[MAXSTR] = "";
	int ret;

	snprintf(pipecommand, MAXSTR, "%s %s \"%s\" \"%s\" 2>&1",
			 GLOBUS_URL_COPY, arguments, src_url, dest_url);
	printf("Executing: %s\n", pipecommand);
	pipe = popen(pipecommand, "r");

	if (!pipe) {
		fprintf(stderr, "Error: couldn't open pipe to %s!\n",
				GLOBUS_URL_COPY);
		return -1;
	}


	while (fgets(linebuf, MAXSTR, pipe))
		fprintf(stdout, ">%s", linebuf);

	ret = pclose(pipe);
	fprintf(stdout, "Command terminated with err_code: %d\n", ret);

	if (ret == 0) {
		return DAP_SUCCESS;
	} else {
		snprintf(error_str, MAXSTR, "%s%s", error_str, linebuf);
		return DAP_ERROR;
	}
}

// Transfer only a part of a file with GUC, given offset and length.
int transfer_guc_part(const char *src, const char *dest, uint8_t streams,
                      uintmax_t off, uintmax_t len, const char *user_args,
                      char *error_str) {
	int status;
	char args[MAXSTR];
	if (len)
		snprintf(args, MAXSTR, "-p %hhu -off %ju -len %ju %s",
		         streams, off, len, user_args);
	else
		snprintf(args, MAXSTR, "-p %hhu -off %ju %s", streams, off, user_args);
	status = transfer_globus_url_copy(src, dest, args, error_str);
	return status;
}
const char *unique_filepath(const char *directory) {
	static char path[_POSIX_PATH_MAX];
	static int unique = 0;
	static time_t curr = 0;

	if (unique) {
		unique++;
	} else {
		unique = 1;
		curr = time(NULL);
	}

	snprintf(path, _POSIX_PATH_MAX, "%s/file-%u-%d-%d-%d", directory,
	         my_ip_addr(), getpid(), (int)curr, unique);
	return path;
}

// For multi-file transfers using a dynamic transfer destination, the list
// specifying the file transfer source-destination pairs must be rewritten:
// all destination URLs must be rewritten with the dynamic destination.
bool translate_file(
		const char *multi_file_xfer_file,  // original static dest url file
		const char *dynamic_multi_file_xfer_file,  // new dynamic dest url file
		const char *dynamic_destination) {  // dynamic dest url to use
	char src[MAXSTR], dest[MAXSTR];
	FILE *static_urls = fopen(multi_file_xfer_file, "r");
	if (!static_urls) {
		fprintf(stderr, "open static URL specification file %s for read: %s\n",
				multi_file_xfer_file, strerror(errno));
		return false;
	}
	FILE *dynamic_urls = fopen(dynamic_multi_file_xfer_file, "w");
	if (!dynamic_urls) {
		fprintf(stderr,
		        "open dynamic URL specification file %s for write: %s\n",
		        dynamic_multi_file_xfer_file, strerror(errno));
		fclose(static_urls);
		return false;
	}

	while (fscanf(static_urls, " %s %s ", src, dest) != EOF)
		fprintf(dynamic_urls, "%s %s\n",
		        src, unique_filepath(dynamic_destination));
	fclose(static_urls);
	fclose(dynamic_urls);
	return true;
}

// Get the size of a file in bytes given a URL. If the size cannot be
// determined (e.g. if the file doesn't exist or the protocol is not
// supported), return SIZE_UNKNOWN.
uintmax_t get_filesize(const char *url) {
	int status;
	char proto[MAXSTR], host[MAXSTR], path[MAXSTR];
	parse_url(url, proto, host, path);

	// If it's a file, use stat64()...
	if (!strcmp(proto, "file")) {
		struct stat64 filestat;
		if (!stat64(path, &filestat))
			return filestat.st_size;
		else
			return SIZE_UNKNOWN;
	}

	// If it's a [gsi]ftp URL, use get_filesize_from_gridftp...
	if (!strcmp(proto, "ftp") || !strcmp(proto, "gsiftp")) {
		uintmax_t filesize;
		status = get_filesize_from_gridftp(url, &filesize);
		if (status != DAP_SUCCESS)
			return SIZE_UNKNOWN;
		return filesize;
	}

	

	// If it's anything else, just say the size is unknown.
	return SIZE_UNKNOWN;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++*/

int do_estimation_service(int argc, char *argv[]) {
	char user_args[MAXSTR];  // GUC args supplied by the user.
	char args[MAXSTR];  // Args which ultimately get passed to GUC.
	char src_url[MAXSTR], dest_url[MAXSTR];
	char error_str[MAXSTR];// disk_transfer[MAXSTR];
	uintmax_t src_filesize, dest_filesize;
	int status;
	struct transfer_status ts;  // Maintains state of the transfer.
	int i;
	bool multi_file_xfer = false;
	char * multi_file_xfer_file = NULL;
	char *file_size, *disk_transfer;
	const char *dynamic_multi_file_xfer_file = NULL;
	bool dynamic_file_xfer = false;
	
	strncpy(src_url, argv[1], MAXSTR);
	strncpy(dest_url, argv[2], MAXSTR);
	ts.src_url = src_url;
	ts.dest_url = dest_url;
	ts.size = 0;
        
	file_size = getenv("FILE_SIZE");
	if (file_size && *file_size) {

		ts.size = filesize_in_bytes(file_size);
	}
	
	// Default to 1GB
	if (!ts.size) ts.size = 1024*1024*1024;
        disk_transfer = getenv("DISK_TRANSFER");
        if (!strcmp(disk_transfer,"YES"))
            ts.size = get_filesize(src_url);
        if (ts.size == 4096)
         ts.size = 1024*1024*1024;
       	ts.off = 0;
		ts.chunk =  1;  // Chunk size
		ts.sample = (ts.size)* 5/100;
		ts.optimized = true;

        for (i = 3; i < argc; i++) {
        if (!strcmp(argv[i], "-s")) {  // Sample size
                if (i+1 >= argc) exit(EXIT_FAILURE);
                ts.sample = filesize_in_bytes(argv[++i]);
            //     printf("Argumets size sampel: %ju\n",ts.sample);
                }
        }
	
       // Debugging
	printf("Source URL: %s\n", ts.src_url);
	printf("Dest URL:   %s\n", ts.dest_url);
	printf("File size:  %ju\n", ts.size);
	//printf("Chunk size is %ju\n", ts.chunk);
//	 printf("Argumets size sampel: %ju\n",ts.sample);
	do_optimization_service(&ts);
	
	//printf("Number of streams: %hhu\n", ts.streams);
	return DAP_SUCCESS;
       }
	
		/* GUC transfer service with optimization***********************/
int do_transfer_service(int argc, char *argv[]) {
	char user_args[MAXSTR];  // GUC args supplied by the user.
	char args[MAXSTR];  // Args which ultimately get passed to GUC.
	char src_url[MAXSTR], dest_url[MAXSTR];
	char error_str[MAXSTR];
	uintmax_t src_filesize, dest_filesize, dest_filesize1;
	int status;
	struct transfer_status ts;  // Maintains state of the transfer.
	int i;
	bool multi_file_xfer = false;
	char *multi_file_xfer_file = NULL;
	const char *dynamic_multi_file_xfer_file = NULL;
	bool dynamic_file_xfer = false;
			
	strncpy(src_url, argv[1], MAXSTR);
	strncpy(dest_url, argv[2], MAXSTR);
	ts.src_url = src_url;
	ts.dest_url = dest_url;
	ts.size = get_filesize(src_url);
	ts.off = 0;
	ts.chunk = 1;//(ts.size-1)/MAX_CHUNKS + 1;  // 
	ts.dynamic = false;
	ts.optimized = false;
        ts.history = false;
	*user_args = '\0';  // Init user_args to empty string 

	for (i = 3; i < argc; i++) {
		if (!strcmp(argv[i], "-s")) {  // Sample size
			if (i+1 >= argc) exit(EXIT_FAILURE);
			ts.sample = filesize_in_bytes(argv[++i]);
			ts.optimized = true;
		} else if (!strcmp(argv[i], "-ch")) {  // Chunk count
			if (i+1 >= argc) exit(EXIT_FAILURE);
			int chunks = atoi(argv[++i]);
			if (chunks < 2) continue;
			ts.dynamic = true;
			ts.chunk = (ts.size-1)/chunks + 1;
		} else if (!strcmp(argv[i], "-h")) {  // History
			ts.history = true;
		} else {
			// Copy anything else into the user_args buffer.
			//snprintf(user_args, MAXSTR, "%s %s", user_args, argv[i]);
			strncat(user_args, " ", MAXSTR);
			strncat(user_args, argv[i], MAXSTR);
		} // Also should take care of -f later.
	}
			
	// Debugging
	printf("Source URL: %s\n", ts.src_url);
	printf("Dest URL:   %s\n", ts.dest_url);
	printf("File size:  %ju\n", ts.size);
//	printf("Chunk size is %ju\n", ts.chunk);

	if (ts.dynamic) printf("Chunk size is %ju\n", ts.chunk);

	if (ts.dynamic && ts.optimized) while (ts.off < ts.size) {
		// Dynamically adjust stream count at every "chunk" boundary.
		// The amount we'll send this time around is min(rem, chunk).
		uintmax_t rem = ts.size - ts.off;
		ts.len = (rem < ts.chunk) ? rem : ts.chunk;

		// Run the optimization service, which will update ts. It will transfer
		// part of the file and may change ts.off, ts.len, and ts.streams.
		do_optimization_service(&ts);

		if (ts.off >= ts.size) break;  // The service transferred everything!

//		printf("Number of streams: %hhu\n", ts.streams);

		// Would it be better to just pass &ts here?
		status = transfer_guc_part(ts.src_url, ts.dest_url,
		                           ts.streams, ts.off, ts.len,
		                           user_args, error_str);

		if (status != DAP_SUCCESS) break;
		
		ts.off += ts.len;
	} else if (ts.optimized) {  // Dynamic optimization is disabled...
		dest_filesize1 = get_filesize(dest_url);

		// Just run the optimization then transfer the file according.
		do_optimization_service(&ts);

		if (ts.off < ts.size)  // The service didn't transfer everything.
//			printf("Number of streams: %hhu\n", ts.streams);

		// Pass zero for len to indicate to just transfer the remainder. If
		// the filesize is known (i.e. the file exists on the remote machine),
		// we need to pass a length of 0 to truncate the file.
		status = transfer_guc_part(ts.src_url, ts.dest_url,
		                           ts.streams,
		                           (dest_filesize1 == SIZE_UNKNOWN) ? ts.off : 0, 0,
		                           user_args, error_str);
	} else {  // No optimization, just transfer the file.
		status = transfer_globus_url_copy(ts.src_url, ts.dest_url,
		                                  user_args, error_str);
	}

	// Check if the transfer failed, and if so report the failure
	if (status == DAP_ERROR) exit(DAP_ERROR);

	// Now verify file size on both ends.
	if (multi_file_xfer) {
		fprintf(stdout, "skipping file size check for multi-file xfer\n");
		fprintf(stderr, "SUCCESS!\n");
		return DAP_SUCCESS;
	}

	src_filesize = ts.size;
	dest_filesize = get_filesize(dest_url);

	if (src_filesize == SIZE_UNKNOWN)
		fprintf(stdout, "size of src (%s) = unknown\n", src_url);
	else
		fprintf(stdout, "size of src (%s) = %ju\n", src_url, src_filesize);

	if (dest_filesize == SIZE_UNKNOWN)
		fprintf(stdout, "size of dest (%s) = unknown\n", dest_url);
	else
		fprintf(stdout, "size of dest (%s) = %ju\n", dest_url, dest_filesize);

	if (src_filesize == SIZE_UNKNOWN || dest_filesize == SIZE_UNKNOWN) {
		fprintf(stderr, "SUCCESS?\n");
		return DAP_SUCCESS;
	}

	if (src_filesize != dest_filesize) {
		fprintf(stderr, "Error: src and dest file sizes do not match!\n");
		return DAP_ERROR;
	}

	fprintf(stderr, "SUCCESS!\n");
	return DAP_SUCCESS;			
}

int main(int argc, char *argv[]) {
	int dap_status;

	if (argc < 3) {
		print_usage(argv[0]);
		exit(-1);
	}

	config();  // read config file

	// TODO(?) The correct parameter name is actually daemonname_ENVIRONMENT, where
	// daemonname is the Stork daemon name, known to the Condor Master, and
	// defined in the DAEMON_LIST parameter.  The daemonname is _usually_ defined
	// to be STORK by convention, but can be anything.
#	define STORK_ENVIRONMENT	"STORK_ENVIRONMENT"
	char* env = param(STORK_ENVIRONMENT);
	if (env) {
		printf("merging environment defined by %s configuration\n",
		       STORK_ENVIRONMENT);
		Env envStrParser;
		// Note: If [name]_ENVIRONMENT is not specified, env will now be null.
		// Env::Merge(null) will always return true, so the warning will not be
		// printed in this case.
		MyString env_errors;
		if (!envStrParser.MergeFromV1RawOrV2Quoted(env, &env_errors)) {
			// this is an invalid env string
			fprintf(stderr, "Warning! Configuration file variable "
					"`%s' has invalid value `%s'; ignoring: %s\n",
					STORK_ENVIRONMENT, env, env_errors.Value());
			free(env);
		} else {
			char **unix_env;
			free(env);
			unix_env = envStrParser.getStringArray();
			for (int j = 0; (unix_env[j] && unix_env[j][0]); j++) {
				if (!SetEnv(unix_env[j])) {
					dprintf(D_ALWAYS, "Failed to put "
					        "\"%s\" into environment.\n", unix_env[j]);
					return DAP_ERROR;
				}
			}
			deleteStringArray(unix_env);
		}
	}
	
	char *dap_type = getenv("DAP_TYPE");

	if (dap_type && !strcmp(dap_type, "estimation")) {
		dap_status=do_estimation_service (argc,argv);
	} else {
		dap_status=do_transfer_service (argc, argv);
	} return dap_status;
}
